(**************************************************************************
* BlockCAD, a program for building with Lego like blocks on the computer
* Copyright (C) 1998  Anders Isaksson
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
* -------------------------------------------------------------------------
* I hope You send me a copy of any important changes/additions You do.
*
* E-mail: isaksson.etuna@ebox.tninet.se
*
**************************************************************************)

unit LcadGnrl;
{
  Handles the new general blocks - polygon definitions read from external files
}

interface

uses
  Windows, Graphics, SysUtils, Classes,
  LCadPc2, IntList, DrawObject
  ;


type

  TPoint3DArray = array[0..100] of TPoint3D;
  PPoint3DArray = ^TPoint3DArray;

  T3DPoly = record
    NumVertex: integer;
    VertexList: PPoint3DArray;
    Border: integer;
    DetLev: integer;
    cl: TColor;
  end;

  tx = class(TBitMap);

  T3DPolyArray = array[0..100] of T3DPoly;
  P3DPolyArray = ^T3DPolyArray;

  TGeneralBlock = class;

  TDrawClass = class(TObject)
  private
    l, b, h: integer;
    fNumPolys: integer;
    Polys: P3DPolyArray;
    DrawOrder: array[0..7] of TIntegerList;
    RotOffset: array[0..3, 0..1] of integer;
    TopStudPoly: integer;
    TopStudPos: PPoint3DArray;
    BtmStudPoly: integer;
    BtmStudCount: integer;
    BtmStudPos: PPoint3DArray;
    fName,
      pName: string;
    fHighLight: integer;
    fLDrawPatch: integer;
    fBitMapFile: string;
    fBitMapPoly: integer;
    fChanged: Boolean;
    constructor CreateFromFile(const FileName, BlockName: string);
    constructor CreateEmpty;
    procedure DeAllocate;
    procedure Draw(const p: TPoint3D; const Dir: integer;
      const DrawCanvas: TDrawObject; const Selected: Boolean);
    procedure TopStuds(const px, py, pz, Dir: integer; const DrawCanvas: TDrawObject);
    procedure DrawTopStud(const px, py, pz: integer; const DrawCanvas: TDrawObject);
    procedure BottomStuds(const px, py, pz, Dir: integer; const DrawCanvas: TDrawObject);
    procedure DrawBtmStud(const px, py, pz: integer; const DrawCanvas: TDrawObject);
    function GetNumPoints(i: integer): integer;
    function GetPolyPoint(i, j: integer): TPoint3D;
    procedure SetPolyPoint(i, j: integer; p: TPoint3D);
    function GetDrawCount(i: integer): integer;
    function GetDrawOrder(i, j: integer): integer;
    procedure SetDrawOrder(i, j, f: integer);
    procedure SetDrawCount(i, c: integer);
    function GetBorder(i: integer): integer;
    procedure SetBorder(i, b: integer);
    function GetColor(i: integer): TColor;
    procedure SetColor(i: integer; b: TColor);
    function GetDetail(i: integer): integer;
    procedure SetDetail(i, b: integer);
    procedure Update(f: TStringList);
  public
    TopStudCount: integer;
    constructor CreateClone(d: TDrawClass);
    destructor Destroy; override;
    procedure Assign(d: TDrawClass);
    property FilNam: string read fName write fName;
    property PrtNam: string read pName write pName;
    property NumPolys: integer read fNumPolys;
    property NumPoints[i: integer]: integer
    read GetNumPoints;
    property Border[i: integer]: integer
    read GetBorder write SetBorder;
    property Color[i: integer]: TColor
    read GetColor write SetColor;
    property Detail[i: integer]: integer
    read GetDetail write SetDetail;
    property PolyPoint[i, j: integer]: TPoint3D
    read GetPolyPoint write SetPolyPoint;
    property HighLight: integer read fHighLight write fHighLight;
    property NumDrawFaces[i: integer]: integer
    read GetDrawCount write SetDrawCount;
    property OrderedFace[i, j: integer]: integer
    read GetDrawOrder write SetDrawOrder;
    property LDrawPatch: integer read fLDrawPatch write fLDrawPatch;
    property Changed: Boolean read fChanged;
    procedure AddPoly(a: array of TPoint3D);
    procedure ClonePoly(ix: integer);
    procedure DeletePoly(i: integer);
    procedure InsertPoint(i, j: integer; p: TPoint3D);
    procedure DeletePoint(i, j: integer);
    procedure GenerateStuds(gl, gb, gh: integer);
    procedure RemoveTopStuds;
    procedure SetLBH(nl, nb, nh: integer);
  end;

  TGeneralBlock = class(TBlock)
  private
    procedure SetDrawClass(const aValue: TDrawClass);
    function GetDrawClass: TDrawClass;
    function GetFileName: string;
    function GetPartName: string;
  public
    constructor Create;
    constructor CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
      const cc: TColor;
      const ss: Boolean;
      const IsBasePlate: Boolean;
      const IsTransp: Boolean;
      const FileName, PartName, LDrawId: string); reintroduce;
    procedure Initialize(const FileName, PartName: string); override;
    procedure Draw(const DrawCanvas: TDrawObject); override;
    procedure TopStuds(const DrawCanvas: TDrawObject); override;
    procedure BottomStuds(const DrawCanvas: TDrawObject); override;
    //    function    HitTest(const xx, yy: integer): Boolean;        override;
    procedure Turn; override;
    function AsString: string; override;
    property FilNam: string read GetFileName;
    property PrtNam: string read GetPartName;
    property DrawClass: TDrawClass read GetDrawClass write SetDrawClass;
    class function CreateFromString(const s: string): TBlock; override;
  end;

procedure UpdateDataFiles;
procedure AddDrawClass(d: TDrawClass);

implementation

uses
  Main, CEJUtils
  //  ,Debug
  ;

var
  RegisteredClasses : TStringList;
  TempList          : TPoint3DArray;

const
  LookUp            : array[0..7, 0..3] of integer = (
    (0, 1, 2, 3),
    (1, 2, 3, 0),
    (2, 3, 0, 1),
    (3, 0, 1, 2),
    (4, 5, 6, 7),
    (5, 6, 7, 4),
    (6, 7, 4, 5),
    (7, 4, 5, 6)
    );

  Turner            : array[0..3, 0..1] of integer = (
    (1, 0),
    (0, 1),
    (-1, 0),
    (0, -1)
    );

  xShift            : array[0..3, 0..1] of integer = (
    (0, 0),
    (0, 0),
    (1, 0),
    (0, 1)
    );

  yShift            : array[0..3, 0..1] of integer = (
    (0, 0),
    (1, 0),
    (0, 1),
    (0, 0)
    );

function Blend(const q1, q2: TRGBQuad; f: TMyFloat): TRGBQuad;
begin
  Result.rgbRed := Round((1 - f) * q1.rgbRed + f * q2.rgbRed);
  Result.rgbGreen := Round((1 - f) * q1.rgbGreen + f * q2.rgbGreen);
  Result.rgbBlue := Round((1 - f) * q1.rgbBlue + f * q2.rgbBlue);
end;

procedure TexMap(Src, Dst: TBitmap; P0, P1, P2: TPoint; Bl: Boolean; Blimit: TMyFloat);
var
  sx, sy, kx, ky    : TMyFloat;
  tx, ty            : integer;
  x, y              : integer;
  xp, yp            : integer;
  w, h              : integer;
  k1, k2            : TMyFloat;
  cl                : TColor;
  xpd, ypd          : TMyFloat;
  d1, d2            : TMyFloat;
  c20x, c10y        : integer;
  x0, x1, y0, y1    : integer;
  P3                : TPoint;
  Trns              : TColor;

  pDst              : PRGBQuadArray;
  dDst              : integer;

  pSrc              : array of PRGBQuadArray;

  q, q1, q2         : TRGBQuad;
  cl1               : TColor;
begin
  w := Src.Width;
  h := Src.Height;

  P3.x := P0.x + (P1.x - P0.x) + (P2.x - P0.x);
  P3.y := P0.y + (P1.y - P0.y) + (P2.y - P0.y);
  c20x := (P2.x - P0.x);
  c10y := (P1.y - P0.y);
  if (c20x = 0) then
    c20x := 1;
  if (c10y = 0) then
    c10y := 1;
  sx := c20x / w;
  sy := c10y / h;
  tx := P0.x;
  ty := P0.y;
  kx := (P2.y - P0.y) / c20x;
  ky := (P1.x - P0.x) / c10y;
  k1 := 1.0 / (sx - kx * ky * sx);
  k2 := 1.0 / (sy - kx * ky * sy);
  x0 := Max(0, Min(P0.x, Min(P1.x, Min(P2.x, P3.x))));
  x1 := Min(Dst.Width - 1, Max(P0.x, Max(P1.x, Max(P2.x, P3.x))));
  y0 := Max(0, Min(P0.y, Min(P1.y, Min(P2.y, P3.y))));
  y1 := Min(Dst.Height - 1, Max(P0.y, Max(P1.y, Max(P2.y, P3.y))));
  Trns := Src.Canvas.Pixels[0, 0];

  dDst := PChar(Dst.ScanLine[1]) - PChar(Dst.ScanLine[0]);
  PChar(pDst) := PChar(Dst.ScanLine[0]) + y0 * dDst;

  SetLength(pSrc, h);

  for y := 0 to Src.Height - 1 do
    pSrc[y] := PRGBQuadArray(Src.ScanLine[y]);

  for y := y0 to y1 do begin
    for x := x0 to x1 do begin
      xpd := (x - tx - ky * (y - ty)) * k1;
      ypd := (y - ty - kx * (x - tx)) * k2;
      xp := Round(xpd);
      yp := Round(ypd);
      if (xp > 0) and (yp > 0) and (xp < Pred(w)) and (yp < Pred(h)) then begin
        d1 := (xpd - xp);
        d2 := (ypd - yp);
        q := pSrc[yp][xp];
        cl := RGB(q.rgbRed, q.rgbGreen, q.rgbBlue);
        if (cl <> Trns) then begin
          if Bl then begin
            q2 := q1;
            if (d1 < -Blimit) then begin
              q1 := pSrc[yp][xp - 1];
              q2 := Blend(q, q1, -d1)
            end
            else if (d1 > Blimit) then begin
              q1 := pSrc[yp][xp + 1];
              q2 := Blend(q, q1, d1);
            end;
            if (d2 < -Blimit) then begin
              q1 := pSrc[yp - 1][xp];
              q2 := Blend(q, q1, -d2);
            end
            else if (d2 > Blimit) then begin
              q1 := pSrc[yp + 1][xp];
              q2 := Blend(q, q1, d2);
            end;
            q := q2;
          end;
          if (x >= 0) then begin
            pDst[x] := q;
          end
        end
      end
    end;
    Inc(PChar(pDst), dDst);
  end
end;

procedure AddDrawClass(d: TDrawClass);
var
  i                 : integer;
begin
  i := RegisteredClasses.IndexOf(d.FilNam + ':' + d.PrtNam);
  if (i < 0) then
    RegisteredClasses.AddObject(d.FilNam + ':' + d.PrtNam, d);
end;

function RegisterDrawClass(FileName, BlockName: string): TDrawClass;
var
  i                 : integer;
begin
  i := RegisteredClasses.IndexOf(FileName + ':' + BlockName);
  if (i >= 0) then
    Result := TDrawClass(RegisteredClasses.Objects[i])
  else begin
    Result := TDrawClass.CreateFromFile(FileName, BlockName);
    RegisteredClasses.AddObject(FileName + ':' + BlockName, Result);
  end;
end;

function LoadPart(FilNam, BlkNam: string; f: TStrings): string;
begin
  Result := BlkNam + '.bcp';
  if FileExists(ExeLib + 'parts\' + Result) then begin
    f.LoadFromFile(ExeLib + 'parts\' + Result);
  end
  else begin
    Result := FilNam;
    if FileExists(ExeLib + 'parts\' + Result) then begin
      f.LoadFromFile(ExeLib + 'parts\' + Result);
    end
    else begin
      Result := '';
      raise Exception.CreateFmt('Could not find part/file %s/%s', [BlkNam, FilNam]);
    end
  end
end;

procedure UpdateDataFiles;
var
  (*  Files:  TStringList; *)
  l                 : TStringList;
  i                 : integer;
  d                 : TDrawClass;
begin
  (*  Files := TStringList.Create; *)
  l := TStringList.Create;
  try
    with RegisteredClasses do begin
      for i := 0 to Pred(Count) do begin
        d := TDrawClass(Objects[i]);
        if d.Changed then begin
          l.Clear;
          d.Update(l);
          l.SaveToFile(ExeLib + 'parts\' + d.pName + '.bcp');
        end
          (*
                  f := Files.IndexOf(d.fName);
                  if (f < 0) then begin
                    l := TStringList.Create;
                    Files.AddObject(d.fName, l);
                    try
                      l.LoadFromFile(ExeLib + '\parts\' + d.fName);
                    except
                      ;
                    end
                  end
                  else begin
                    l := TStringList(Files.Objects[f]);
                  end;
                  d.Update(l);
          *)
      end
    end;
  finally
    (*
        for i := 0 to Pred(Files.Count) do begin
          TStringList(Files.Objects[i]).SaveToFile(ExeLib + '\parts\' + Files[i]);
          Files.Objects[i].Free;
        end;
        Files.Free;
    *)
    l.Free;
  end;
end;

(*************************************************************************)

constructor TDrawClass.CreateEmpty;
begin
  inherited Create;
  HighLight := -1;
end;

constructor TDrawClass.CreateClone(d: TDrawClass);
begin
  inherited Create;
  Assign(d);
end;

procedure TDrawClass.Assign(d: TDrawClass);
var
  i                 : integer;
  j                 : integer;
  dc                : integer;
  pp                : P3dPolyArray;
  pv                : PPoint3DArray;
begin
  fName := D.fName;
  pName := D.pName;
  fNumPolys := D.NumPolys;
  fHighLight := D.fHighLight;
  GetMem(pp, NumPolys * SizeOf(T3DPoly));
  Polys := pp;
  l := D.l;
  b := D.b;
  h := D.h;
  for i := 0 to Pred(NumPolys) do begin
    Polys[i].NumVertex := D.Polys[i].NumVertex;
    Polys[i].Border := D.Polys[i].Border;
    Polys[i].DetLev := D.Polys[i].DetLev;
    Polys[i].cl := D.Polys[i].cl;
    GetMem(pv, Polys[i].NumVertex * SizeOf(TPoint3D));
    Polys[i].VertexList := pv;
    for j := 0 to Pred(Polys[i].NumVertex) do begin
      pv[j].x := D.Polys[i].VertexList[j].x;
      pv[j].y := D.Polys[i].VertexList[j].y;
      pv[j].z := D.Polys[i].VertexList[j].z;
    end;
  end;
  { copy drawing order per octant 'Qx' }
  for i := 0 to 7 do begin
    dc := D.DrawOrder[i].Count;
    DrawOrder[i] := TIntegerList.Create;
    for j := 0 to Pred(dc) do begin
      DrawOrder[i][j] := D.DrawOrder[i][j];
    end;
  end;
  { copy topstuds info 'TS' }
  TopStudPoly := D.TopStudPoly;
  TopStudCount := D.TopStudCount;
  if TopStudCount > 0 then begin
    GetMem(pv, TopStudCount * SizeOf(TPoint3D));
    TopStudPos := pv;
    for i := 0 to Pred(TopStudCount) do begin
      pv[i].x := D.TopStudPos[i].x;
      pv[i].y := D.TopStudPos[i].y;
      pv[i].z := D.TopStudPos[i].z;
    end;
  end;
  { copy bottomstuds info 'BS' }
  BtmStudPoly := D.BtmStudPoly;
  BtmStudCount := D.BtmStudCount;
  if BtmStudCount > 0 then begin
    GetMem(pv, BtmStudCount * SizeOf(TPoint3D));
    BtmStudPos := pv;
    for i := 0 to Pred(BtmStudCount) do begin
      pv[i].x := D.BtmStudPos[i].x;
      pv[i].y := D.BtmStudPos[i].y;
      pv[i].z := Polys[BtmStudPoly].VertexList[0].z;
    end;
  end;
  { copy color index for the individual polygons 'Cx' }
  { copy LDrawPatch }
  fLDrawPatch := D.fLDrawPatch;
  fBitMapFile := D.fBitMapFile;
  fBitMapPoly := D.fBitMapPoly;
  RotOffset := D.RotOffset;
  fChanged := D.fChanged;
end;

constructor TDrawClass.CreateFromFile(const FileName, BlockName: string);
var
  f                 : TStringList;
  i                 : integer;
  j                 : integer;
  dc                : integer;
  pp                : P3dPolyArray;
  pv                : PPoint3DArray;
  s                 : string;
  cl                : TColor;

const
  Buf               : string = '';
  BufP              : integer = 0;

  procedure First(s: string);
  begin
    Buf := s;
    BufP := 1;
  end;

  procedure SkipBlanks;
  begin
    while (BufP <= Length(Buf)) and (Buf[BufP] in [#9, ' ', ',', ';']) do
      Inc(BufP);
  end;

  function NextString: string;
  var
    i               : integer;
  begin
    SkipBlanks;
    i := BufP;
    while (BufP <= Length(Buf)) and not (Buf[BufP] = ',') do
      Inc(BufP);
    Result := Copy(Buf, i, BufP - i);
  end;

  function NextInt: integer;
  var
    i               : integer;
  begin
    SkipBlanks;
    i := BufP;
    if (BufP <= Length(Buf)) and (Buf[BufP] = '-') then
      Inc(BufP);
    while (BufP <= Length(Buf)) and (Buf[BufP] in ['0'..'9']) do
      Inc(BufP);
    if ((BufP - i) > 0) then
      Result := StrToInt(Copy(Buf, i, BufP - i))
    else
      Result := 0;
  end;

  function FirstInt(s: string): integer;
  begin
    First(s);
    Result := NextInt;
  end;

  function CurChar: char;
  begin
    if Bufp <= Length(Buf) then
      Result := Buf[Bufp]
    else
      Result := #0;
  end;

  procedure LocatePart(FilNam, BlkNam: string);
  var
    i, j            : integer;
  begin
    i := Pos('.dat', LowerCase(BlkNam));
    if (i > 0) and (i = (Length(BlkNam) - 3)) then begin
      BlkNam := Copy(BlkNam, 1, i - 1);
    end;
    LoadPart(FilNam, BlkNam, f);
    i := f.IndexOf('[' + BlkNam + ']');
    if (i < 0) then
      raise exception.CreateFmt('Could not find part %s in file %s', [BlkNam, FilNam]);
    while (i >= 0) do begin
      f.Delete(i);
      Dec(i);
    end;
    i := 0;
    while (i < f.Count) and (Length(f[i]) > 0) and (f[i][1] <> '[') do
      Inc(i);
    for j := Pred(f.Count) downto i do
      f.Delete(j);
  end;

begin
  CreateEmpty;
  f := TStringList.Create;
  fName := FileName;
  pName := BlockName;
  try
    LocatePart(FileName, BlockName);
    fNumPolys := StrToInt(f.Values['NP']);
    GetMem(pp, NumPolys * SizeOf(T3DPoly));
    Polys := pp;
    s := f.Values['BB'];
    l := FirstInt(s);
    b := NextInt;
    h := NextInt;
    for i := 0 to Pred(NumPolys) do begin
      s := f.Values[Format('P%d', [i + 1])];
      Polys[i].NumVertex := FirstInt(s);
      Polys[i].Border := NextInt;
      GetMem(pv, Polys[i].NumVertex * SizeOf(TPoint3D));
      Polys[i].VertexList := pv;
      for j := 0 to Pred(Polys[i].NumVertex) do begin
        pv[j].x := NextInt;
        pv[j].y := NextInt;
        pv[j].z := NextInt;
      end;
      Polys[i].DetLev := NextInt;       { Optional Detail level }
      cl := NextInt;
      if (cl = -1) then
        Polys[i].cl := MainForm.ColorMatch2(NextInt) { Optional fixed color  }
      else
        Polys[i].cl := cl;
    end;
    { read drawing order per octant 'Qx' }
    for i := 0 to 7 do begin
      s := f.Values[Format('Q%d', [i + 1])];
      dc := FirstInt(s);
      DrawOrder[i] := TIntegerList.Create;
      for j := 0 to Pred(dc) do begin
        DrawOrder[i][j] := Pred(NextInt);
      end;
    end;
    { read topstuds info 'TS' }
    s := f.Values['TS'];
    if (Length(s) > 8) then begin
      TopStudPoly := Pred(FirstInt(s));
      TopStudCount := NextInt;
      GetMem(pv, TopStudCount * SizeOf(TPoint3D));
      TopStudPos := pv;
      for i := 0 to Pred(TopStudCount) do begin
        pv[i].x := NextInt;
        pv[i].y := NextInt;
        if (CurChar = ';') then begin
          pv[i].z := NextInt;
        end
        else
          pv[i].z := Polys[TopStudPoly].VertexList[0].z;
      end;
    end;
    { read bottomstuds info 'BS' }
    s := f.Values['BS'];
    if (Length(s) > 8) then begin
      BtmStudPoly := Pred(FirstInt(s));
      BtmStudCount := NextInt;
      GetMem(pv, BtmStudCount * SizeOf(TPoint3D));
      BtmStudPos := pv;
      for i := 0 to Pred(BtmStudCount) do begin
        pv[i].x := NextInt;
        pv[i].y := NextInt;
        if (CurChar = ';') then begin
          pv[i].z := NextInt;
        end
        else
          pv[i].z := Polys[BtmStudPoly].VertexList[0].z;
      end;
    end;
    { read LDrawPatch - optional }
    s := f.Values['LD'];
    if (Length(s) > 0) then try
      fLDrawPatch := StrToInt(s);
    except
      fLDrawPatch := 0;
    end;
    { read bitmappoly/file - optional }
    s := f.Values['BM'];
    if (Length(s) > 0) then begin
      fBitMapPoly := Pred(FirstInt(s));
      fBitMapFile := NextString;
    end
    else begin
      fBitMapPoly := -1;
    end;
    { read rot.offsets - optional }
    s := f.Values['RO'];
    if (Length(s) > 0) then begin
      RotOffset[0, 0] := FirstInt(s);
      RotOffset[0, 1] := NextInt;
      RotOffset[1, 0] := NextInt;
      RotOffset[1, 1] := NextInt;
      RotOffset[2, 0] := NextInt;
      RotOffset[2, 1] := NextInt;
      RotOffset[3, 0] := NextInt;
      RotOffset[3, 1] := NextInt;
    end
    else begin
      RotOffset[0, 0] := 0;
      RotOffset[0, 1] := 0;
      RotOffset[1, 0] := 0;
      RotOffset[1, 1] := 0;
      RotOffset[2, 0] := 0;
      RotOffset[2, 1] := 0;
      RotOffset[3, 0] := 0;
      RotOffset[3, 1] := 0;
    end;
  finally
    f.Free;
  end;
  fChanged := False;
end;

procedure TDrawClass.Update(f: TStringList);
var
  i                 : integer;
  j                 : integer;
  s                 : string;

begin
  { Remove old definition }
  i := f.IndexOf('[' + pName + ']');
  if (i >= 0) then begin
    f.Delete(i);
    while (i < f.Count) and (Length(f[i]) > 0) and (f[i][1] <> '[') do begin
      f.Delete(i);
    end
  end;
  { Write polygons }
  f.Add(Format('[%s]', [pName]));
  f.Add(Format('NP=%d', [fNumPolys]));
  f.Add(Format('BB=%d,%d,%d', [l, b, h]));
  if (LDrawPatch <> 0) then
    f.Add(Format('LD=%d', [LDrawPatch]));
  for i := 0 to Pred(NumPolys) do begin
    s := Format('P%d=%d, %d', [i + 1, Polys[i].NumVertex, Polys[i].Border]);
    for j := 0 to Pred(Polys[i].NumVertex) do begin
      s := s + Format(', %d,%d,%d', [Polys[i].VertexList[j].x,
        Polys[i].VertexList[j].y,
          Polys[i].VertexList[j].z]);
    end;
    { Write detail level - optional }
    if (Polys[i].DetLev > 0) or (Polys[i].cl <> 0) then
      s := s + Format(', %d', [Polys[i].DetLev]);
    { Write fixed color - optional }
    if (Polys[i].cl <> 0) then
      s := s + Format(', %d', [Polys[i].cl]);
    f.Add(s);
  end;
  { Write bitmappoly/file - optional }
  if (fBitMapPoly > -1) then begin
    f.Add(Format('BM=%d, %s', [Succ(fBitMapPoly), fBitMapFile]));
  end;
  { Write drawing order per octant 'Qx' }
  for i := 0 to 7 do begin
    s := Format('Q%d=%d ', [Succ(i), DrawOrder[i].Count]);
    for j := 0 to Pred(DrawOrder[i].Count) do begin
      s := s + Format(',%d', [Succ(DrawOrder[i][j])]);
    end;
    f.Add(s);
  end;
  { Write topstuds info 'TS' }
  s := Format('TS=%d,%d', [Succ(TopStudPoly), TopStudCount]);
  for i := 0 to Pred(TopStudCount) do begin
    s := s + Format(', %d,%d', [TopStudPos[i].x, TopStudPos[i].y]);
    if (TopStudPos[i].z <> Polys[TopStudPoly].VertexList[0].z) then
      s := s + Format(';%d', [TopStudPos[i].z]);
  end;
  f.Add(s);
  { Write bottomstuds info 'BS' }
  s := Format('BS=%d,%d', [Succ(BtmStudPoly), BtmStudCount]);
  for i := 0 to Pred(BtmStudCount) do begin
    s := s + Format(', %d,%d', [BtmStudPos[i].x, BtmStudPos[i].y]);
    if (BtmStudPos[i].z <> Polys[BtmStudPoly].VertexList[0].z) then
      s := s + Format(';%d', [BtmStudPos[i].z]);
  end;
  f.Add(s);
  { Write rotation offsets 'RO' - optional}
  for i := 0 to 3 do begin
    if (RotOffset[i, 0] <> 0) or (RotOffset[i, 1] <> 0) then begin
      s := Format('RO=%d,%d, %d,%d, %d,%d, %d,%d',
        [RotOffset[0, 0], RotOffset[0, 1],
        RotOffset[1, 0], RotOffset[1, 1],
          RotOffset[2, 0], RotOffset[2, 1],
          RotOffset[3, 0], RotOffset[3, 1]
          ]);
      f.Add(s);
      Break;
    end
  end;
  fChanged := False;
end;

procedure TDrawClass.DeAllocate;
var
  i                 : integer;
begin
  if NumPolys > 0 then begin            { Deallocate dynamic memory }
    for i := 0 to Pred(NumPolys) do begin
      if Polys[i].NumVertex > 0 then begin
        FreeMem(Polys[i].VertexList, Polys[i].NumVertex * SizeOf(TPoint3D));
        Polys[i].VertexList := nil;
      end;
    end;
    FreeMem(Polys, NumPolys * SizeOf(T3DPoly));
  end;
  for i := 0 to 7 do begin
    if DrawOrder[i] <> nil then begin
      DrawOrder[i].Free;
      DrawOrder[i] := nil;
    end;
  end;
  if (TopStudPoly >= 0) and Assigned(TopStudPos) then begin
    FreeMem(TopStudPos, TopStudCount * SizeOf(TPoint3D));
    TopStudPos := nil;
    TopStudCount := 0;
  end;
  if (BtmStudPoly >= 0) and Assigned(BtmStudPos) then begin
    FreeMem(BtmStudPos, BtmStudCount * SizeOf(TPoint3D));
    BtmStudPos := nil;
    BtmStudCount := 0;
  end;
  fName := '';
  pName := '';
  fBitMapFile := '';

end;

destructor TDrawClass.Destroy;
begin
  DeAllocate;
  inherited Destroy;
end;

function TDrawClass.GetNumPoints(i: integer): integer;
begin
  if (i >= 0) and (i < NumPolys) then begin
    Result := Polys[i].NumVertex;
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

function TDrawClass.GetBorder(i: integer): integer;
begin
  if (i >= 0) and (i < NumPolys) then begin
    Result := Polys[i].Border;
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

procedure TDrawClass.SetBorder(i, b: integer);
begin
  if (i >= 0) and (i < NumPolys) then begin
    Polys[i].Border := b;
    fChanged := True;
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

function TDrawClass.GetColor(i: integer): TColor;
begin
  if (i >= 0) and (i < NumPolys) then begin
    Result := Polys[i].cl;
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

procedure TDrawClass.SetColor(i: integer; b: TColor);
begin
  if (i >= 0) and (i < NumPolys) then begin
    Polys[i].cl := b;
    fChanged := True;
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

function TDrawClass.GetDetail(i: integer): integer;
begin
  if (i >= 0) and (i < NumPolys) then begin
    Result := Polys[i].DetLev;
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

procedure TDrawClass.SetDetail(i, b: integer);
begin
  if (i >= 0) and (i < NumPolys) then begin
    Polys[i].DetLev := b;
    fChanged := True;
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

function TDrawClass.GetDrawCount(i: integer): integer;
begin
  if (i >= 0) and (i <= 7) and Assigned(DrawOrder[i]) then begin
    Result := DrawOrder[i].Count;
  end
  else begin
    raise Exception.Create('Illegal octant for GetDrawCount');
  end;
end;

procedure TDrawClass.SetDrawCount(i, c: integer);
begin
  if (i >= 0) and (i <= 7) then begin
    while (DrawOrder[i].Count > c) do
      DrawOrder[i].Delete(Pred(DrawOrder[i].Count));
    fChanged := True;
  end
  else begin
    raise Exception.Create('Illegal octant for SetDrawCount');
  end;
end;

function TDrawClass.GetDrawOrder(i, j: integer): integer;
begin
  if (i >= 0) and (i <= 7) and Assigned(DrawOrder[i]) then begin
    if (j >= 0) and (j < DrawOrder[i].Count) then begin
      Result := DrawOrder[i][j];
    end
    else begin
      raise Exception.Create('Illegal face index for GetDrawOrder');
    end;
  end
  else begin
    raise Exception.Create('Illegal octant for GetDrawOrder');
  end;
end;

procedure TDrawClass.SetDrawOrder(i, j, f: integer);
begin
  if (i >= 0) and (i <= 7) and Assigned(DrawOrder[i]) then begin
    if (j >= 0) then begin
      DrawOrder[i][j] := f;
      fChanged := True;
    end
    else begin
      raise Exception.Create('Illegal face index for SetDrawOrder');
    end;
  end
  else begin
    raise Exception.Create('Illegal octant for SetDrawOrder');
  end;
end;

function TDrawClass.GetPolyPoint(i, j: integer): TPoint3D;
begin
  if (i >= 0) and (i < NumPolys) then begin
    if (j >= 0) and (j < Polys[i].NumVertex) then begin
      Result := Polys[i].VertexList[j];
    end
    else begin
      raise Exception.Create('Illegal point index');
    end
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

procedure TDrawClass.SetPolyPoint(i, j: integer; p: TPoint3D);
begin
  if (i >= 0) and (i < NumPolys) then begin
    if (j >= 0) and (j < Polys[i].NumVertex) then begin
      Polys[i].VertexList[j] := p;
      fChanged := True;
    end
    else begin
      raise Exception.Create('Illegal point index');
    end
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
end;

procedure TDrawClass.AddPoly(a: array of TPoint3D);
var
  i, j              : integer;
begin
  ReallocMem(Polys, (NumPolys + 1) * SizeOf(T3DPoly));
  Polys[NumPolys].NumVertex := High(a) - Low(a) + 1;
  Polys[NumPolys].Border := 1;
  Polys[NumPolys].DetLev := 0;
  Polys[NumPolys].cl := 0;
  GetMem(Polys[NumPolys].VertexList, Polys[NumPolys].NumVertex * SizeOf(TPoint3D));
  j := 0;
  for i := Low(a) to High(a) do begin
    Polys[NumPolys].VertexList[j] := a[i];
    Inc(j);
  end;
  for i := 0 to 7 do begin
    DrawOrder[i].Add(NumPolys);
  end;
  Inc(fNumPolys);
  fChanged := True;
end;

procedure TDrawClass.ClonePoly(ix: integer);
var
  i                 : integer;
begin
  ReallocMem(Polys, (NumPolys + 1) * SizeOf(T3DPoly));
  Polys[NumPolys].NumVertex := Polys[ix].NumVertex;
  Polys[NumPolys].Border := Polys[ix].Border;
  Polys[NumPolys].DetLev := Polys[ix].DetLev;
  Polys[NumPolys].cl := Polys[ix].cl;
  GetMem(Polys[NumPolys].VertexList, Polys[NumPolys].NumVertex * SizeOf(TPoint3D));
  for i := 0 to Pred(Polys[ix].NumVertex) do begin
    Polys[NumPolys].VertexList[i] := Polys[ix].VertexList[i];
  end;
  for i := 0 to 7 do begin
    DrawOrder[i].Add(NumPolys);
  end;
  Inc(fNumPolys);
  fChanged := True;
end;

procedure TDrawClass.DeletePoly(i: integer);
var
  j, k              : integer;
begin
  { Remove this polygon from all draw orders    }
  { and renumber the polys with higher number   }
  for j := 0 to 7 do begin
    DrawOrder[j].Remove(i);
    for k := 0 to Pred(DrawOrder[j].Count) do begin
      if (DrawOrder[j][k] > i) then
        DrawOrder[j][k] := Pred(DrawOrder[j][k]);
    end;
  end;
  { If this is the top or bottom stud polygon,  }
  { remove those also                           }
  if (i = TopStudPoly) then begin
    TopStudPoly := -1;
    FreeMem(TopStudPos, TopStudCount * SizeOf(TPoint3D));
    TopStudCount := 0;
    TopStudPos := nil;
  end
  else if (i = BtmStudPoly) then begin
    BtmStudPoly := -1;
    FreeMem(BtmStudPos, BtmStudCount * SizeOf(TPoint3D));
    BtmStudCount := 0;
    BtmStudPos := nil;
  end;
  { Renumber TopStudPoly and BtmStudPoly ?      }
  if TopStudPoly > i then
    Dec(TopStudPoly);
  if BtmStudPoly > i then
    Dec(BtmStudPoly);
  { Lastly, remove the polygon                  }
  if Polys[i].NumVertex > 0 then begin
    FreeMem(Polys[i].VertexList, Polys[i].NumVertex * SizeOf(TPoint3D));
    Polys[i].VertexList := nil;
  end;
  for j := i to NumPolys - 2 do begin
    Polys[j] := Polys[j + 1];
  end;
  Dec(fNumPolys);
  fChanged := True;
  { Here we could reallocate also, but we don't }
end;

procedure TDrawClass.InsertPoint(i, j: integer; p: TPoint3D);
var
  k                 : integer;
begin
  if (i >= 0) and (i < NumPolys) then begin
    if (j >= 0) and (j < Polys[i].NumVertex) then begin
      Inc(Polys[i].NumVertex);
      ReallocMem(Polys[i].VertexList, Polys[i].NumVertex * SizeOf(TPoint3D));
      for k := Pred(Polys[i].NumVertex) downto j + 1 do
        Polys[i].VertexList[k] := Polys[i].VertexList[k - 1];
      Polys[i].VertexList[j] := p;
    end
    else begin
      raise Exception.Create('Illegal point index');
    end
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
  fChanged := True;
end;

procedure TDrawClass.DeletePoint(i, j: integer);
var
  k                 : integer;
begin
  if (i >= 0) and (i < NumPolys) then begin
    if (j >= 0) and (j < Polys[i].NumVertex) then begin
      for k := j to Pred(Pred(Polys[i].NumVertex)) do
        Polys[i].VertexList[k] := Polys[i].VertexList[k + 1];
      Dec(Polys[i].NumVertex);
      ReallocMem(Polys[i].VertexList, Polys[i].NumVertex * SizeOf(TPoint3D));
    end
    else begin
      raise Exception.Create('Illegal point index');
    end
  end
  else begin
    raise Exception.Create('Illegal polygon index');
  end;
  fChanged := True;
end;

procedure TDrawClass.GenerateStuds(gl, gb, gh: integer);
var
  i, j, k, c        : integer;
  xx, yy            : integer;
  Rgn               : HRgn;

  procedure FindTopPoly;
  var
    p, i, j, zm     : integer;
  label
    NextPoly;
  begin
    zm := -9999;
    j := -1;
    for p := 0 to Pred(NumPolys) do begin
      if Polys[p].VertexList[0].z > zm then begin
        for i := 1 to Pred(Polys[p].NumVertex) do begin
          if Polys[p].VertexList[i].z <> Polys[p].VertexList[0].z then
            goto NextPoly;
        end;
        zm := Polys[p].VertexList[0].z;
        j := p;
        NextPoly:
      end
    end;
    if j >= 0 then
      TopStudPoly := j;
  end;

  procedure FindBtmPoly;
  var
    p, i, j, zm     : integer;
  label
    NextPoly;
  begin
    zm := 9999;
    j := -1;
    for p := 0 to Pred(NumPolys) do begin
      if Polys[p].VertexList[0].z < zm then begin
        for i := 1 to Pred(Polys[p].NumVertex) do begin
          if Polys[p].VertexList[i].z <> Polys[p].VertexList[0].z then
            goto NextPoly;
        end;
        zm := Polys[p].VertexList[0].z;
        j := p;
        NextPoly:
      end
    end;
    if j >= 0 then
      BtmStudPoly := j;
  end;

  procedure PreparePolygonRegion(p: integer; var a: array of TPoint);
  var
    i               : integer;
  begin
    for i := 0 to Pred(Polys[p].NumVertex) do begin
      a[i].x := Polys[p].VertexList[i].x;
      a[i].y := Polys[p].VertexList[i].y;
    end;
  end;

begin
  c := Max(0, gl * gb);
  if (TopStudPoly < 0) then
    Exit;
  FindTopPoly;
  if true {(c <> TopStudCount)} then begin
    PreparePolygonRegion(TopStudPoly, Pts100);
    Rgn := CreatePolygonRgn(Pts100, Polys[TopStudPoly].NumVertex, ALTERNATE);
    try
      ReallocMem(TopStudPos, c * SizeOf(TPoint3D));
      i := 0;
      for j := 0 to Pred(gl) do begin
        for k := 0 to Pred(gb) do begin
          xx := (PFACTORXY div 2) + (j * PFACTORXY);
          yy := (PFACTORXY div 2) + (k * PFACTORXY);
          if PtInRegion(Rgn, xx, yy) then begin
            if PtInRegion(Rgn, xx + (PFACTORXY div 4) + 2, yy) and
              PtInRegion(Rgn, xx - (PFACTORXY div 4) - 2, yy) and
              PtInRegion(Rgn, xx, yy + (PFACTORXY div 4) + 2) and
              PtInRegion(Rgn, xx, yy - (PFACTORXY div 4) - 2) then begin
              TopStudPos[i].x := xx;
              TopStudPos[i].y := yy;
              TopStudPos[i].z := Polys[TopStudPoly].VertexList[0].z;
              Inc(i);
            end
            else begin
              Dec(c);
            end
          end
          else begin
            Dec(c);
          end
        end
      end;
      ReallocMem(TopStudPos, c * SizeOf(TPoint3D));
      TopStudCount := c;
    finally
      DeleteObject(Rgn);
    end;
    fChanged := True;
  end;
  c := Max(0, (gl - 1) * (gb - 1));
  if (BtmStudPoly < 0) then
    Exit;
  FindBtmPoly;
  if true {(c <> BtmStudCount)} then begin
    PreparePolygonRegion(BtmStudPoly, Pts100);
    Rgn := CreatePolygonRgn(Pts100, Polys[BtmStudPoly].NumVertex, ALTERNATE);
    try
      ReallocMem(BtmStudPos, c * SizeOf(TPoint3D));
      i := 0;
      for j := 1 to Pred(gl) do begin
        for k := 1 to Pred(gb) do begin
          xx := (PFACTORXY div 2) + (j * PFACTORXY);
          yy := (PFACTORXY div 2) + (k * PFACTORXY);
          if PtInRegion(Rgn, xx, yy) then begin
            if PtInRegion(Rgn, xx + (PFACTORXY div 4) + 2, yy) and
              PtInRegion(Rgn, xx - (PFACTORXY div 4) - 2, yy) and
              PtInRegion(Rgn, xx, yy + (PFACTORXY div 4) + 2) and
              PtInRegion(Rgn, xx, yy - (PFACTORXY div 4) - 2) then begin
              BtmStudPos[i].x := (j * PFACTORXY);
              BtmStudPos[i].y := (k * PFACTORXY);
              BtmStudPos[i].z := Polys[BtmStudPoly].VertexList[0].z;
              Inc(i);
            end
            else begin
              Dec(c);
            end
          end
          else begin
            Dec(c);
          end
        end
      end;
      ReallocMem(BtmStudPos, c * SizeOf(TPoint3D));
      BtmStudCount := c;
    finally
      DeleteObject(Rgn);
    end;
    fChanged := True;
  end;
end;

procedure TDrawClass.RemoveTopStuds;
begin
  FreeMem(TopStudPos, TopStudCount * SizeOf(TPoint3D));
  TopStudCount := 0;
  TopStudPos := nil;
  TopStudCount := 0;
end;

procedure TDrawClass.SetLBH(nl, nb, nh: integer);
begin
  l := nl;
  b := nb;
  h := nh;
  fChanged := True;
end;

procedure TDrawClass.Draw(const p: TPoint3D; const Dir: integer;
  const DrawCanvas: TDrawObject; const Selected: Boolean);

  procedure Turn(const Source: array of TPoint3D; var Dest: array of TPoint3D; const Dir: integer);
  var
    i               : integer;
    x, y            : integer;
  begin
    for i := Low(Source) to High(Source) do begin
      x := Source[i].x;
      y := Source[i].y;
      Dest[i].x := x * Turner[Dir][0] + y * Turner[Dir][1] +
        l * xShift[Dir][0] + b * xShift[Dir][1];
      Dest[i].y := y * Turner[Dir][0] - x * Turner[Dir][1] +
        l * yShift[Dir][0] + b * yShift[Dir][1];
      Dest[i].z := Source[i].z;
    end;
  end;

  procedure Translate(const Source: array of TPoint3D; var Dest: array of TPoint3D; const Amount: TPoint3D);
  var
    i               : integer;
  begin
    for i := Low(Source) to High(Source) do begin
      Dest[i].x := Source[i].x + p.x;
      Dest[i].y := Source[i].y + p.y;
      Dest[i].z := Source[i].z + p.z;
    end;
  end;

  procedure DrawPoly(i: integer);
  var
    cp, cb          : TColor;
    up, bc          : TColor;
    bs              : TBrushStyle;
    bmp             : TBitMap;
    bmp1            : TBitMap;
    zpl             : ZPLANE;
    f, t1,
      t2            : TMyFloat;
    RGB             : TColor;
    R, G, B         : integer;
    c               : TColor;
    k, k1           : integer;
    fl              : integer;
    bl              : integer;
    DrawRect        : TRect;
  begin
    if (Polys[i].DetLev <= DetailLevel) then begin

      Turn(Slice(Polys[i].VertexList^, Polys[i].NumVertex), TempList, Dir);
      Translate(Slice(TempList, Polys[i].NumVertex), TempList, p);
      Transform(Slice(TempList, Polys[i].NumVertex), Pts100, 0);

      cp := DrawCanvas.Canvas.Pen.Color;
      cb := DrawCanvas.Canvas.Brush.Color;
      bc := cb;
      bs := DrawCanvas.Canvas.Brush.Style;

      if (i = HighLight) then begin
        DrawCanvas.Canvas.Pen.Color := SelEdge;
        bc := SelColor;
        DrawCanvas.Canvas.Brush.Color := bc;
        DrawCanvas.Canvas.Brush.Style := bsSolid;
      end
      else if (Polys[i].cl <> 0) then begin
        if Polys[i].cl < 0 then begin
          bc := -Polys[i].cl;
          DrawCanvas.Canvas.Brush.Color := bc;
          DrawCanvas.Canvas.Brush.Style := bsDiagCross;
        end
        else begin
          bc := Polys[i].cl;
          DrawCanvas.Canvas.Brush.Color := bc;
        end;
        if bs = bsClear then begin
          DrawCanvas.Canvas.Pen.Color := bc;
          DrawCanvas.Canvas.Brush.Style := bsClear;
        end;
      end;

      (* Calculate direction of plane, and shade accordingly *)
      if (DoShade) then begin
        ShadeEdge(bc, DrawCanvas.Canvas.Pen.Color, EdgeShade, c);
        if not Selected then
          DrawCanvas.Canvas.Pen.Color := c;
        if Polys[i].NumVertex > 2 then begin
          calc_zplane(zpl, TempList[0], TempList[1], TempList[2]);
          t1 := sqrt(light.x * light.x + light.y * light.y + light.z * light.z);

          t2 := sqrt(zpl.A * zpl.A + zpl.B * zpl.B + zpl.C * zpl.C);
          if (t1 <> 0) and (t2 <> 0) then begin
            f := (light.x * zpl.A + light.y * zpl.B - light.z * zpl.C) / t1 / t2;
            // reduce the effect of shading -- we don't want parts of the image
            // to be completely blackened out.
(*
  When _all_ parts have been created as 'general' and checked for CCW, the abs()
  can be taken out and the factor (and threshold) can be adjusted to give a more
  real looking shading (backside darker than frontside). Lots of work...
*)
            f := (abs(f / 4)) + Threshold;
            if ({(f < Threshold) or}(f > 1.0)) then begin
              // should never happen
              f := 1.0;
            end;
            RGB := ColorToRGB(DrawCanvas.Canvas.Brush.Color);
            B := Round(f * (RGB and $FF0000)) and $FF0000; {}
            R := Round(f * (RGB and $00FF00)) and $00FF00; {}
            G := Round(f * (RGB and $0000FF)) and $0000FF; {}
            DrawCanvas.Canvas.Brush.Color := B or R or G;
          end
        end
      end;

      (* --------------------------------------------------- *)

      if DrawCanvas.Canvas.Brush.Style = bsClear then
        bl := 1
      else
        bl := Polys[i].Border;

      case bl of
        0: begin                        {No border lines at all}
            DrawCanvas.Canvas.Pen.Color := DrawCanvas.Canvas.Brush.Color; //bc;
            DrawCanvas.Polygon(Slice(Pts100, Polys[i].NumVertex));
          end;
        1: begin                        {All border lines}
            DrawCanvas.Polygon(Slice(Pts100, Polys[i].NumVertex));
          end;
        else begin                      {Some border lines}
            up := DrawCanvas.Canvas.Pen.Color;
            DrawCanvas.Canvas.Pen.Color := DrawCanvas.Canvas.Brush.Color;
            DrawCanvas.Polygon(Slice(Pts100, Polys[i].NumVertex));
            DrawCanvas.Canvas.Pen.Color := up;
            k := 0;
            fl := 2;
            while k < Min(31, Polys[i].NumVertex) do begin
              if (Polys[i].Border and fl) = 0 then begin
                k1 := k + 1;
                if k1 >= Polys[i].NumVertex then
                  k1 := 0;
                DrawCanvas.MoveTo(Pts100[k].x, Pts100[k].y);
                DrawCanvas.LineTo(Pts100[k1].x, Pts100[k1].y);
              end;
              Inc(k);
              fl := fl shl 1;
            end;
          end
      end;

      DrawCanvas.Canvas.Pen.Color := cp;
      DrawCanvas.Canvas.Brush.Color := cb;
      DrawCanvas.Canvas.Brush.Style := bs;

      if (i = fBitMapPoly) and (i <> HighLight) then begin
        if (DetailLevel > 1) then begin
          DrawRect.Left := 9999;
          DrawRect.Top := 9999;
          DrawRect.Right := -9999;
          DrawRect.Bottom := -9999;
          for k := 0 to Pred(Polys[i].NumVertex) do begin
            DrawRect.Left := Min(Pts100[k].x, DrawRect.Left);
            DrawRect.Top := Min(Pts100[k].y, DrawRect.Top);
            DrawRect.Right := Max(Pts100[k].x, DrawRect.Right);
            DrawRect.Bottom := Max(Pts100[k].y, DrawRect.Bottom);
          end;
          if (DrawRect.Bottom >= 0) and
            (DrawRect.Right >= 0) and
            ((DrawRect.Right - DrawRect.Left) > 0) and
            ((DrawRect.Bottom - DrawRect.Top) > 0) then begin
            bmp := TBitMap.Create;
            bmp1 := TBitMap.Create;
            bmp1.PixelFormat := pf32bit;
            try
              bmp1.Height := DrawRect.Bottom + 1;
              bmp1.Width := DrawRect.Right + 1;
              bmp1.Canvas.CopyRect(DrawRect, DrawCanvas.Canvas, DrawRect);
              try
                Bmp.LoadFromFile(ExeLib + 'parts\' + fBitMapFile);
                bmp.PixelFormat := pf32bit;
                TexMap(Bmp, Bmp1, Pts100[0], Pts100[1], Pts100[3], True, 0.0);
                DrawCanvas.CopyRect(DrawRect, bmp1.Canvas, DrawRect);
              except
                ;
              end;
            finally
              Bmp.Free;
              Bmp1.Free;
            end
          end
        end
        else begin                      { Just draw an X }
          DrawCanvas.MoveTo(Pts100[0].x, Pts100[0].y);
          DrawCanvas.LineTo(Pts100[2].x, Pts100[2].y);
          DrawCanvas.MoveTo(Pts100[1].x, Pts100[1].y);
          DrawCanvas.LineTo(Pts100[3].x, Pts100[3].y);
        end
      end
    end
  end;

var
  i                 : integer;

begin                                   {TDrawClass.Draw}
  for i := 0 to Pred(DrawOrder[LookUp[Octant, Dir]].Count) do begin
    DrawPoly(DrawOrder[LookUp[Octant, Dir]][i]);
  end;
end;

procedure TDrawClass.DrawTopStud(const px, py, pz: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(px, py, pz)], Pts4, 0);
  with Pts4[0] do begin
    StudSwitcher(x, y, DrawCanvas);
  end;
end;

procedure TDrawClass.DrawBtmStud(const px, py, pz: integer; const DrawCanvas: TDrawObject);
begin
  Transform([Point3D(px, py, pz)], Pts4, 0);
  with Pts4[0] do begin
    DrawCanvas.PolyLine([Point(x - xf, y + yf2),
      Point(x - xf, y - yf2),
        Point(x - xf2, y - yf),
        Point(x + xf2, y - yf),
        Point(x + xf, y - yf2),
        Point(x + xf, y + yf2),
        Point(x + xf2, y + yf),
        Point(x - xf2, y + yf),
        Point(x - xf, y + yf2)]);
  end;
end;

procedure TDrawClass.TopStuds(const px, py, pz, Dir: integer; const DrawCanvas: TDrawObject);
var
  i                 : integer;
  //  shz               : TColor;
  //  c                 : TColor;
begin
  if (TopStudCount > 0) and Assigned(TopStudPos) then begin
    //    if (MainForm.miShading.Checked) then begin
          //      ShadeEdge(DrawCanvas.Canvas.Brush.Color, DrawCanvas.Canvas.Pen.Color, EdgeShade, c);
          //      DrawCanvas.Canvas.Pen.Color := c;
          //      Shade(DrawCanvas.Canvas.Brush.Color, lfz, shz);
          //      DrawCanvas.Canvas.Brush.Color := shz;
    //    end;
    for i := 0 to Pred(TopStudCount) do
      DrawTopStud(TopStudPos[i].x * Turner[Dir][0] + TopStudPos[i].y * Turner[Dir][1] +
        l * xShift[Dir][0] + b * xShift[Dir][1] + px,
        TopStudPos[i].y * Turner[Dir][0] - TopStudPos[i].x * Turner[Dir][1] +
        l * yShift[Dir][0] + b * yShift[Dir][1] + py,
        TopStudPos[i].z + pz,
        DrawCanvas)
  end;
end;

procedure TDrawClass.BottomStuds(const px, py, pz, Dir: integer; const DrawCanvas: TDrawObject);
var
  i                 : integer;
  shz               : TColor;
  c                 : TColor;
begin
  if (BtmStudCount > 0) and Assigned(BtmStudPos) then begin
    if (MainForm.miShading.Checked) then begin
      ShadeEdge(DrawCanvas.Canvas.Brush.Color, DrawCanvas.Canvas.Pen.Color, EdgeShade, c);
      DrawCanvas.Canvas.Pen.Color := c;
      Shade(DrawCanvas.Canvas.Brush.Color, lfz, shz);
      DrawCanvas.Canvas.Brush.Color := shz;
    end;
    for i := 0 to Pred(BtmStudCount) do
      DrawBtmStud(BtmStudPos[i].x * Turner[Dir][0] + BtmStudPos[i].y * Turner[Dir][1] +
        l * xShift[Dir][0] + b * xShift[Dir][1] + px,
        BtmStudPos[i].y * Turner[Dir][0] - BtmStudPos[i].x * Turner[Dir][1] +
        l * yShift[Dir][0] + b * yShift[Dir][1] + py,
        BtmStudPos[i].z + pz,
        DrawCanvas)
  end;
end;

(*************************************************************************)

constructor TGeneralBlock.CreateBlock(const xx, yy, zz, ll, bb, hh: integer;
  const cc: TColor;
  const ss: Boolean;
  const IsBasePlate: Boolean;
  const IsTransp: Boolean;
  const FileName, PartName, LDrawId: string);
begin
  inherited CreateBlock(xx, yy, zz, ll, bb, hh, cc, ss, IsBasePlate, IsTransp, LDrawId);
  Initialize(FileName, PartName);
  LDrawPatch := DrawClass.LDrawPatch;
end;

constructor TGeneralBlock.Create;
begin
  inherited Create;
end;

procedure TGeneralBlock.SetDrawClass(const aValue: TDrawClass);
begin
  gpPointer := Pointer(aValue);
  LDrawPatch := DrawClass.LDrawPatch;
end;

function TGeneralBlock.GetDrawClass: TDrawClass;
begin
  Result := TDrawClass(gpPointer);
end;

procedure TGeneralBlock.Initialize(const FileName, PartName: string);
var
  dl, db            : integer;
begin
  DrawClass := RegisterDrawClass(FileName, PartName);
  LDrawPatch := DrawClass.LDrawPatch;
  if Odd(Dir) then begin
    db := DrawClass.l;
    dl := DrawClass.b;
  end
  else begin
    dl := DrawClass.l;
    db := DrawClass.b;
  end;
  pl := dl;
  pb := db;
  ph := DrawClass.h;
  t := 'G';
end;

procedure TGeneralBlock.Draw(const DrawCanvas: TDrawObject);
var
  p                 : TPoint3D;
begin
  if (pl = 0) then Exit;

  DoShade := MainForm.miShading.Checked;

  if not Selected then begin
    DrawCanvas.Canvas.Brush.Color := cl;
    DrawCanvas.Canvas.Pen.Color := EdgeColor(cl);
  end
  else begin
    DrawCanvas.Canvas.Brush.Color := SelColor;
    DrawCanvas.Canvas.Brush.Color := cl;
    if cl = RedColor then
      DrawCanvas.Canvas.Pen.Color := clWhite
    else
      DrawCanvas.Canvas.Pen.Color := RedColor;
  end;

  if (bp and MainForm.WireBase and (Octant > 3)) or
    (MainForm.WireFrame) then begin
    DrawCanvas.Canvas.Brush.Style := bsClear; { Draw as Wire }
    DrawCanvas.Canvas.Pen.Color := cl;  {}
    DoShade := False;
  end
  else if (cl = clAqua) or (Transp) then begin
    DrawCanvas.Canvas.Brush.Style := bsDiagCross; { Transparent piece }
  end;

  p.x := px;
  p.y := py;
  p.z := pz;
  DrawClass.Draw(p, Dir, DrawCanvas, Selected);

  if (st and not MainForm.btnStuds.Down) then
    if Octant < 4 then begin
      if (TopCover < pl * pb) then
        TopStuds(DrawCanvas);
    end
    else if (BotCover < pl * pb) then
      BottomStuds(DrawCanvas);

  DrawCanvas.Canvas.Brush.Color := clSilver;
  DrawCanvas.Canvas.Brush.Style := bsSolid;
  DrawCanvas.Canvas.Pen.Color := BlackColor;

end;

procedure TGeneralBlock.TopStuds(const DrawCanvas: TDrawObject);
var
  c                 : TColor;
begin
  if (MainForm.miShading.Checked) and not Selected then begin
    ShadeEdge(DrawCanvas.Canvas.Brush.Color, DrawCanvas.Canvas.Pen.Color, EdgeShade, c);
    DrawCanvas.Canvas.Pen.Color := c;
    Shade(DrawCanvas.Canvas.Brush.Color, lfz, shz);
    DrawCanvas.Canvas.Brush.Color := shz;
  end;
  DrawClass.TopStuds(px, py, pz, Dir, DrawCanvas);
end;

procedure TGeneralBlock.BottomStuds(const DrawCanvas: TDrawObject);
begin
  DrawClass.BottomStuds(px, py, pz, Dir, DrawCanvas);
end;
(*
function TGeneralBlock.HitTest(const xx, yy: integer): Boolean;
var Rgn: HRgn;

  procedure PrepareRegion(const px, py, pz, pl ,pb, ph: integer);
  begin
    Transform([Point3D(px,py,pz),         Point3D(px,py,pz+ph),
               Point3D(px+pl,py,pz+ph),   Point3D(px+pl,py+pb,pz+ph),
               Point3D(px+pl,py+pb,pz),   Point3D(px,py+pb,pz)],
               Pts6, 0);
  end;

begin
  Result := False;
  PrepareRegion(px, py, pz, pl, pb, ph);
  Rgn := CreatePolygonRgn(Pts6, 6, ALTERNATE);
  if (Rgn <> 0) then begin
    try
      Result := PtInRegion(Rgn, xx, yy);
    finally
      DeleteObject(Rgn);
    end
  end
end;
*)

procedure TGeneralBlock.Turn;
var
  rox, roy          : integer;
begin
  rox := DrawClass.RotOffset[Dir, 0];
  roy := DrawClass.RotOffset[Dir, 1];
  px := px + rox * PFACTORXY;
  py := py + roy * PFACTORXY;
  inherited Turn;
end;

function TGeneralBLock.GetFileName: string;
begin
  if Assigned(DrawClass) then
    Result := DrawClass.FilNam
  else
    Result := '';
end;

function TGeneralBLock.GetPartName: string;
begin
  if Assigned(DrawClass) then
    Result := DrawClass.PrtNam
  else
    Result := '';
end;

function TGeneralBlock.AsString: string;
begin
  if t = 'G' then begin
    Result := Format('%s:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%d:%s:%s:%d:%s:%d',
      [t, px, py, pz, pl, pb, ph, cl,
      integer(st), integer(bp), Dir, integer(Transp),
        FilNam, PrtNam, Group, PartID, Step]);
  end
  else
    Result := inherited AsString;
end;

class function TGeneralBlock.CreateFromString(const s: string): TBlock;
var
  a                 : TStringList;
  Res               : TGeneralBlock;
begin
  if s[1] = 'G' then begin
    Res := TGeneralBlock.Create;
    a := TStringList.Create;
    try
      if (Split(s, a) <> 17) then
        raise Exception.Create('Illegal string to CreateFromString: ' + s);
      Res.t := a[0][1];
      Res.px := StrToInt(a[1]);
      Res.py := StrToInt(a[2]);
      Res.pz := StrToInt(a[3]);
      Res.pl := StrToInt(a[4]);
      Res.pb := StrToInt(a[5]);
      Res.ph := StrToInt(a[6]);
      Res.cl := StrToInt(a[7]);
      Res.st := a[8] = '1';
      Res.bp := a[9] = '1';
      Res.Dir := StrToInt(a[10]);
      Res.Transp := a[11] = '1';
      Res.Initialize(a[12], a[13]);
      Res.Group := StrToInt(a[14]);
      Res.Partid := a[15];
      Res.Step := StrToInt(a[16]);
    finally
      a.Free;
    end;
    Result := Res;
  end
  else begin
    Result := inherited CreateFromString(s);
  end;
end;

(*************************************************************************)

procedure CleanUp; far
var
  i                 : integer;
begin
  for i := 0 to Pred(RegisteredClasses.Count) do begin
    RegisteredClasses.Objects[i].Free;
  end;
  RegisteredClasses.Free;
end;

initialization
  AddExitProc(CleanUp);
  RegisteredClasses := TStringList.Create;
  RegisteredClasses.Sorted := True;
  RegisteredClasses.Duplicates := dupError;
end.

